@@ -2,7 +2,7 @@ import os
 from spb.defaults import cfg
 from spb.backends.base_backend import Plot
 from sympy.external import import_module
-
+import warnings
 
 def compute_streamlines(x, y, u, v, density=1.0):
     """Return streamlines of a vector flow.
@@ -193,8 +193,12 @@ class BokehBackend(Plot):
         contours, quivers, streamlines...
         To learn more about customization:
 
-        * Refer to [#fn1]_ to customize lines plots. Default to:
-          ``dict(line_width = 2)``.
+        * Refer to:
+
+          - [#fn1]_ to customize lines plots. Default to:
+            ``dict(line_width = 2)``.
+          - [#fn6]_ to customize scatter plots. Default to:
+            ``dict(marker = "circle")``.
 
         * Default options for quiver plots:
 
@@ -218,11 +222,38 @@ class BokehBackend(Plot):
         If True, the backend will update the data series over the visibile
         range whenever a pan-event is triggered. Default to True.
 
+    annotations : list, optional
+        A list of dictionaries specifying the type of annotation
+        required. The keys in the dictionary should be equivalent
+        to the arguments of the `bokeh.models.LabelSet` class.
+        This feature is experimental. It might get removed in the future.
+
+    markers : list, optional
+        A list of dictionaries specifying the type the markers required.
+        The keys in the dictionary should be equivalent to the arguments
+        of the `bokeh.models.Scatter` class.
+        This feature is experimental. It might get removed in the future.
+
+    rectangles : list, optional
+        A list of dictionaries specifying the dimensions of the
+        rectangles to be plotted. The ``"args"`` key must contain the
+        `bokeh.models.ColumnDataSource` object containing the
+        data. All other keyword arguments will be passed to the
+        `bokeh.models.Rect` class.
+        This feature is experimental. It might get removed in the future.
+
+    fill : dict, optional
+        A dictionary specifying the type of color filling required in
+        the plot. The keys in the dictionary should be equivalent to the
+        arguments of the `bokeh.models.VArea` class.
+        This feature is experimental. It might get removed in the future.
+
 
     References
     ==========
-    .. [#fn1] https://docs.bokeh.org/en/latest/docs/reference/plotting.html?highlight=line#bokeh.plotting.Figure.line
+    .. [#fn1] https://docs.bokeh.org/en/latest/docs/reference/plotting.html#bokeh.plotting.Figure.line
     .. [#fn2] https://docs.bokeh.org/en/latest/docs/reference/themes.html
+    .. [#fn6] https://docs.bokeh.org/en/latest/docs/reference/plotting/figure.html#bokeh.plotting.Figure.scatter
 
 
     See also
@@ -232,6 +263,8 @@ class BokehBackend(Plot):
     """
 
     _library = "bokeh"
+    _allowed_keys = Plot._allowed_keys + [
+        "markers", "annotations", "fill", "rectangles", "update_event"]
 
     colorloop = []
     colormaps = []
@@ -244,6 +277,7 @@ class BokehBackend(Plot):
         self.bokeh = import_module(
             'bokeh',
             import_kwargs={'fromlist': ['models', 'events', 'plotting', 'io', 'palettes', 'embed', 'resources']},
+            warn_not_installed=True,
             min_module_version='2.3.0')
         bp = self.bokeh.palettes
         cc = import_module(
@@ -322,7 +356,8 @@ class BokehBackend(Plot):
             self._fig.grid.minor_grid_line_alpha = cfg["bokeh"]["minor_grid_line_alpha"]
             self._fig.grid.minor_grid_line_color = self._fig.grid.grid_line_color[0]
             self._fig.grid.minor_grid_line_dash = cfg["bokeh"]["minor_grid_line_dash"]
-        self._fig.on_event(self.bokeh.events.PanEnd, self._pan_update)
+        if self._update_event:
+            self._fig.on_event(self.bokeh.events.PanEnd, self._pan_update)
 
     @property
     def fig(self):
@@ -373,7 +408,8 @@ class BokehBackend(Plot):
                         else next(self._cm)
                     )
                     ds, line, cb, kw = self._create_gradient_line(
-                        x, y, param, colormap, s.get_label(self._use_latex), s.rendering_kw)
+                        x, y, param, colormap, s.get_label(self._use_latex),
+                        s.rendering_kw, s.is_point)
                     self._fig.add_glyph(ds, line)
                     if self.legend:
                         self._handles[i] = cb
@@ -398,10 +434,11 @@ class BokehBackend(Plot):
                         self._fig.line("xs", "ys", source=source, **kw)
                     else:
                         lkw["size"] = 8
+                        lkw["marker"] = "circle"
                         if not s.is_filled:
                             lkw["fill_color"] = "white"
                         kw = merge({}, lkw, s.rendering_kw)
-                        self._fig.circle("xs", "ys", source=source, **kw)
+                        self._fig.scatter("xs", "ys", source=source, **kw)
 
             elif s.is_contour and (not s.is_complex):
                 if s.is_polar:
@@ -415,6 +452,9 @@ class BokehBackend(Plot):
                 ckw = dict(palette=cm)
                 kw = merge({}, ckw, s.rendering_kw)
 
+                if not s.is_filled:
+                    warnings.warn("Bokeh does not support line contours.")
+
                 self._fig.image(
                     image=[z],
                     x=minx,
@@ -444,16 +484,25 @@ class BokehBackend(Plot):
                     self._fig.multi_line(xs, ys, **kw)
                 else:
                     x, y, u, v = s.get_data()
+                    mag = np.sqrt(u**2 + v**2)
+                    u0, v0 = [t.copy() for t in [u, v]]
+                    if s.normalize:
+                        u, v = [t / mag for t in [u, v]]
                     data, quiver_kw = self._get_quivers_data(x, y, u, v,
                         **s.rendering_kw.copy())
-                    mag = data["magnitude"]
+                    color_val = mag
+                    if s.color_func is not None:
+                        color_val = s.color_func(x, y, u0, v0)
+                    color_val = np.tile(color_val.flatten(), 3)
+                    data["color_val"] = color_val
 
                     color_mapper = self.bokeh.models.LinearColorMapper(
-                        palette=next(self._cm), low=min(mag), high=max(mag))
+                        palette=next(self._cm),
+                        low=min(color_val), high=max(color_val))
                     # don't use color map if a scalar field is visible or if
                     # use_cm=False
                     line_color = (
-                        {"field": "magnitude", "transform": color_mapper}
+                        {"field": "color_val", "transform": color_mapper}
                         if ((not s.use_quiver_solid_color) and s.use_cm)
                         else next(self._cl)
                     )
@@ -510,6 +559,21 @@ class BokehBackend(Plot):
                 kw = merge({}, pkw, s.rendering_kw)
                 self._fig.patch(x, y, **kw)
 
+            elif s.is_generic:
+                if s.type == "markers":
+                    kw = merge({}, {"color": next(self._cl)}, s.rendering_kw)
+                    self._fig.scatter(*s.args, **kw)
+                elif s.type == "annotations":
+                    self._fig.add_layout(
+                        self.bokeh.models.LabelSet(*s.args, **s.rendering_kw))
+                elif s.type == "fill":
+                    kw = merge({}, {"fill_color": next(self._cl)}, s.rendering_kw)
+                    self._fig.varea(*s.args, **kw)
+                elif s.type == "rectangles":
+                    kw = merge({}, {"fill_color": next(self._cl)}, s.rendering_kw)
+                    glyph = self.bokeh.models.Rect(**kw)
+                    self._fig.add_glyph(*s.args, glyph)
+
             else:
                 raise NotImplementedError(
                     "{} is not supported by {}\n".format(type(s), type(self).__name__)
@@ -562,9 +626,12 @@ class BokehBackend(Plot):
         us = u[:-1]
         return xs, ys, us
 
-    def _create_gradient_line(self, x, y, u, colormap, name, line_kw):
+    def _create_gradient_line(self, x, y, u, colormap, name, line_kw, is_point=False):
         merge = self.merge
-        xs, ys, us = self._get_segments(x, y, u)
+        if not is_point:
+            xs, ys, us = self._get_segments(x, y, u)
+        else:
+            xs, ys, us = x, y, u
         color_mapper = self.bokeh.models.LinearColorMapper(
             palette=colormap, low=min(us), high=max(us))
         data_source = self.bokeh.models.ColumnDataSource(
@@ -576,7 +643,10 @@ class BokehBackend(Plot):
             line_color={"field": "us", "transform": color_mapper},
         )
         kw = merge({}, lkw, line_kw)
-        glyph = self.bokeh.models.MultiLine(xs="xs", ys="ys", **kw)
+        if not is_point:
+            glyph = self.bokeh.models.MultiLine(xs="xs", ys="ys", **kw)
+        else:
+            glyph = self.bokeh.models.Scatter(x="xs", y="ys", **kw)
         colorbar = self.bokeh.models.ColorBar(
             color_mapper=color_mapper, title=name, width=8)
         return data_source, glyph, colorbar, kw
@@ -594,7 +664,10 @@ class BokehBackend(Plot):
 
                 if s.is_2Dline and s.is_parametric and s.use_cm:
                     x, y, param = self.series[i].get_data()
-                    xs, ys, us = self._get_segments(x, y, param)
+                    if not s.is_point:
+                        xs, ys, us = self._get_segments(x, y, param)
+                    else:
+                        xs, ys, us = x, y, param
                     rend[i].data_source.data.update({"xs": xs, "ys": ys, "us": us})
                     if i in self._handles.keys():
                         cb = self._handles[i]
@@ -634,27 +707,39 @@ class BokehBackend(Plot):
                         rend[i].data_source.data.update({"xs": xs, "ys": ys})
                     else:
                         quiver_kw = s.rendering_kw.copy()
+                        mag = np.sqrt(u**2 + v**2)
+                        u0, v0 = [t.copy() for t in [u, v]]
+                        if s.normalize:
+                            u, v = [t / mag for t in [u, v]]
                         data, quiver_kw = self._get_quivers_data(
                             x, y, u, v, **quiver_kw
                         )
+                        color_val = mag
+                        if s.color_func is not None:
+                            color_val = s.color_func(x, y, u0, v0)
+                        color_val = np.tile(color_val.flatten(), 3)
+                        data["color_val"] = color_val
                         rend[i].data_source.data.update(data)
 
                         line_color = rend[i].glyph.line_color
                         if (not s.use_quiver_solid_color) and s.use_cm:
                             # update the colorbar
                             cmap = line_color["transform"].palette
-                            mag = data["magnitude"]
+                            color_val = data["color_val"]
                             color_mapper = self.bokeh.models.LinearColorMapper(
-                                palette=cmap, low=min(mag),
-                                high=max(mag))
+                                palette=cmap, low=min(color_val),
+                                high=max(color_val))
                             line_color = quiver_kw.get(
                                 "line_color",
                                 {
-                                    "field": "magnitude",
+                                    "field": "color_val",
                                     "transform": color_mapper
                                 },
                             )
                             rend[i].glyph.line_color = line_color
+                            cb = self._handles[i]
+                            cb.color_mapper.update(
+                                low=min(color_val), high=max(color_val))
 
                 elif s.is_complex and s.is_domain_coloring and not s.is_3Dsurface:
                     # TODO: for some unkown reason, domain_coloring and
@@ -699,7 +784,7 @@ class BokehBackend(Plot):
         merge = self.merge
 
         ext = os.path.splitext(path)[1]
-        if ext.lower() in [".html", ".html"]:
+        if ext.lower() in [".htm", ".html"]:
             CDN = self.bokeh.resources.CDN
             file_html = self.bokeh.embed.file_html
             skw = dict(resources=CDN, title="Bokeh Plot")
@@ -815,7 +900,6 @@ class BokehBackend(Plot):
             "x1": x1s,
             "y0": y0s,
             "y1": y1s,
-            "magnitude": np.tile(magnitude, 3),
         }
 
         return data, quiver_kw
