@@ -2,6 +2,7 @@ import os
 from spb.defaults import cfg
 from spb.backends.base_backend import Plot
 from spb.backends.utils import compute_streamtubes
+from spb.series import PlaneSeries
 from spb.utils import get_vertices_indices
 from sympy.external import import_module
 import warnings
@@ -29,7 +30,7 @@ class K3DBackend(Plot):
 
           - ``scale``: a float number acting as a scale multiplier.
           - ``pivot``: indicates the part of the arrow that is anchored to the
-            X, Y, Z grid. It can be ``"tail", "mid", "middle", "head"``.
+            X, Y, Z grid. It can be ``"tail", "mid", "middle", "tip"``.
           - ``color``: set a solid color by specifying an integer color. If this
             key is not provided, a default color or colormap is used, depenging
             on the value of ``use_cm``.
@@ -82,12 +83,16 @@ class K3DBackend(Plot):
 
     _library = "k3d"
 
+    wireframe_color = 0x000000
     colormaps = []
     cyclic_colormaps = []
+    skip_notebook_check = False
 
     # quivers-pivot offsets
     _qp_offset = {"tail": 0, "mid": 0.5, "middle": 0.5, "tip": 1}
 
+    _allowed_keys = Plot._allowed_keys + ["show_label"]
+
     def __new__(cls, *args, **kwargs):
         return object.__new__(cls)
 
@@ -95,14 +100,17 @@ class K3DBackend(Plot):
         self.k3d = k3d = import_module(
             'k3d',
             import_kwargs={'fromlist': ['helpers', 'objects']},
+            warn_not_installed=True,
             min_module_version='2.9.7')
         cc = import_module(
             'colorcet',
+            warn_not_installed=True,
             min_module_version='3.0.0')
         self.matplotlib = import_module(
             'matplotlib',
             import_kwargs={'fromlist': ['tri', 'cm']},
             min_module_version='1.1.0',
+            warn_not_installed=True,
             catch=(RuntimeError,))
         cm = self.matplotlib.cm
 
@@ -121,9 +129,9 @@ class K3DBackend(Plot):
 
         self._init_cyclers()
         super().__init__(*args, **kwargs)
-        if self._get_mode() != 0:
-            raise ValueError(
-                "Sorry, K3D backend only works within Jupyter Notebook")
+        if (not self.skip_notebook_check) and (self._get_mode() != 0):
+            warnings.warn(
+                "K3DBackend only works properly within Jupyter Notebook")
         self._use_latex = kwargs.get("use_latex", cfg["k3d"]["use_latex"])
         self._set_labels("%s")
 
@@ -133,19 +141,25 @@ class K3DBackend(Plot):
         self._handles = dict()
         self.grid = kwargs.get("grid", cfg["k3d"]["grid"])
 
-        self._fig = k3d.plot(
+        kw = dict(
             grid_visible=self.grid,
             menu_visibility=True,
-            background_color=int(cfg["k3d"]["bg_color"]),
-            grid_color=int(cfg["k3d"]["grid_color"]),
-            label_color=int(cfg["k3d"]["label_color"]),
         )
+        if cfg["k3d"]["bg_color"]:
+            kw["background_color"] = int(cfg["k3d"]["bg_color"])
+        if cfg["k3d"]["grid_color"]:
+            kw["grid_color"] = int(cfg["k3d"]["grid_color"])
+        if cfg["k3d"]["label_color"]:
+            kw["label_color"] = int(cfg["k3d"]["label_color"])
+        if cfg["k3d"]["camera_mode"]:
+            kw["camera_mode"] = cfg["k3d"]["camera_mode"]
+
+        self._fig = k3d.plot(**kw)
         if (self.xscale == "log") or (self.yscale == "log"):
             warnings.warn(
                 "K3D-Jupyter doesn't support log scales. We will "
                 + "continue with linear scales."
             )
-        self.plot_shown = False
 
     @property
     def fig(self):
@@ -167,10 +181,6 @@ class K3DBackend(Plot):
         # show=False
         self._process_series(self._series)
 
-    def _set_piecewise_color(self, s, color):
-        """Set the color to the given series"""
-        raise NotImplementedError
-
     @staticmethod
     def _do_sum_kwargs(p1, p2):
         kw = p1._copy_kwargs()
@@ -218,28 +228,42 @@ class K3DBackend(Plot):
 
         for ii, s in enumerate(series):
             if s.is_3Dline and s.is_point:
-                x, y, z, _ = s.get_data()
+                if s.is_parametric:
+                    x, y, z, param = s.get_data()
+                else:
+                    x, y, z = s.get_data()
+                    param = np.zeros_like(x)
                 positions = np.vstack([x, y, z]).T.astype(np.float32)
-                a = dict(point_size=0.2, color=self._convert_to_int(next(self._cl)))
+                a = dict(point_size=0.2,
+                    color=self._convert_to_int(next(self._cl)))
+                if s.use_cm:
+                    a["color_map"] = next(self._cm)
+                    a["attribute"] = param
                 kw = merge({}, a, s.rendering_kw)
                 plt_points = self.k3d.points(positions=positions, **kw)
                 plt_points.shader = "mesh"
                 self._fig += plt_points
 
             elif s.is_3Dline:
-                x, y, z, param = s.get_data()
+                if s.is_parametric:
+                    x, y, z, param = s.get_data()
+                else:
+                    x, y, z = s.get_data()
+                    param = np.zeros_like(x)
                 vertices = np.vstack([x, y, z]).T.astype(np.float32)
                 # keyword arguments for the line object
                 a = dict(
-                    width=0.1,
+                    width=0.1 if s.show_in_legend else 0.001,
                     name=self._get_series_label(s, "%s") if self._show_label else None,
-                    color=self._convert_to_int(next(self._cl)) if s.line_color is None else s.line_color,
+                    color=(
+                        self.wireframe_color if not s.show_in_legend
+                        else (self._convert_to_int(next(self._cl)) if s.line_color is None else s.line_color)),
                     shader="mesh",
                 )
                 if s.use_cm:
                     a["attribute"] = (param.astype(np.float32),)
                     a["color_map"] = next(self._cm)
-                    a["color_range"] = [param.min(), param.max()]
+                    a["color_range"] = [float(param.min()), float(param.max())]
                 kw = merge({}, a, s.rendering_kw)
                 line = self.k3d.line(vertices, **kw)
                 self._fig += line
@@ -252,14 +276,19 @@ class K3DBackend(Plot):
                     attribute = s.eval_color_func(vertices[:, 0], vertices[:, 1], vertices[:, 2], u.flatten().astype(np.float32), v.flatten().astype(np.float32))
                 else:
                     x, y, z = s.get_data()
-                    x = x.flatten()
-                    y = y.flatten()
-                    z = z.flatten()
-                    vertices = np.vstack([x, y, z]).T.astype(np.float32)
-                    indices = Triangulation(x, y).triangles.astype(np.uint32)
+                    if isinstance(s, PlaneSeries):
+                        # avoid triangulation errors when plotting vertical
+                        # planes
+                        vertices, indices = get_vertices_indices(x, y, z)
+                    else:
+                        x = x.flatten()
+                        y = y.flatten()
+                        z = z.flatten()
+                        vertices = np.vstack([x, y, z]).T.astype(np.float32)
+                        indices = Triangulation(x, y).triangles.astype(np.uint32)
                     attribute = s.eval_color_func(vertices[:, 0], vertices[:, 1], vertices[:, 2])
 
-                self._high_aspect_ratio(x, y, z)
+                # self._high_aspect_ratio(x, y, z)
                 a = dict(
                     name=s.get_label(self._use_latex, "%s") if self._show_label else None,
                     side="double",
@@ -270,7 +299,11 @@ class K3DBackend(Plot):
                 if s.use_cm:
                     a["color_map"] = next(self._cm)
                     a["attribute"] = attribute
-                    a["color_range"] = [attribute.min(), attribute.max()]
+                    # NOTE: color_range must contains elements of type float.
+                    # If np.float32 is provided, mgspack will fail to serialize
+                    # it, hence no html export, hence no screenshots on
+                    # documentation.
+                    a["color_range"] = [float(attribute.min()), float(attribute.max())]
 
                 kw = merge({}, a, s.rendering_kw)
                 surf = self.k3d.mesh(vertices, indices, **kw)
@@ -291,21 +324,21 @@ class K3DBackend(Plot):
                     color=self._convert_to_int(next(self._cl))
                 )
                 kw = merge({}, a, s.rendering_kw)
-                plt_iso = self.k3d.marching_cubes(r.astype(np.float32), **kw)
+                plt_iso = self.k3d.marching_cubes(r.T.astype(np.float32), **kw)
 
                 self._fig += plt_iso
 
             elif s.is_3Dvector and s.is_streamlines:
                 xx, yy, zz, uu, vv, ww = s.get_data()
-                vertices, magn = compute_streamtubes(
-                    xx, yy, zz, uu, vv, ww, s.rendering_kw)
+                vertices, color_val = compute_streamtubes(
+                    xx, yy, zz, uu, vv, ww, s.rendering_kw, s.color_func)
 
                 stream_kw = s.rendering_kw.copy()
                 skw = dict(width=0.1, shader="mesh")
                 if s.use_cm and ("color" not in stream_kw.keys()):
                     skw["color_map"] = next(self._cm)
-                    skw["color_range"] = [np.nanmin(magn), np.nanmax(magn)]
-                    skw["attribute"] = magn
+                    skw["color_range"] = [float(np.nanmin(color_val)), float(np.nanmax(color_val))]
+                    skw["attribute"] = color_val
                 else:
                     col = stream_kw.pop("color", next(self._cl))
                     if not isinstance(col, int):
@@ -333,7 +366,7 @@ class K3DBackend(Plot):
                     solid_color = col * np.ones(xx.size)
                     self._handles[ii] = [qkw, colormap]
 
-                origins, vectors, colors = self._build_k3d_vector_data(xx, yy, zz, uu, vv, ww, qkw, colormap)
+                origins, vectors, colors = self._build_k3d_vector_data(xx, yy, zz, uu, vv, ww, qkw, colormap, s.normalize, s.color_func)
                 if colors is None:
                     colors = solid_color
                 vec_colors = self._create_vector_colors(colors)
@@ -405,9 +438,13 @@ class K3DBackend(Plot):
             self._fig += self.k3d.text2d(
                 self.title, position=[0.025, 0.015], color=0, size=1, label_box=False
             )
+
+        self._new_camera_position()
+        self._add_clipping_planes()
         self._fig.auto_rendering = True
 
-    def _build_k3d_vector_data(self, xx, yy, zz, uu, vv, ww, qkw, colormap):
+    def _build_k3d_vector_data(self, xx, yy, zz, uu, vv, ww, qkw, colormap,
+        normalize, color_func):
         """Assemble the origins, vectors and colors (if possible) matrices.
         """
         np = import_module('numpy')
@@ -417,13 +454,18 @@ class K3DBackend(Plot):
         ]
         scale = qkw["scale"]
         magnitude = np.sqrt(uu ** 2 + vv ** 2 + ww ** 2)
+        if normalize:
+            uu, vv, ww = [t / magnitude for t in [uu, vv, ww]]
         vectors = np.array((uu, vv, ww)).T * scale
         origins = np.array((xx, yy, zz)).T
 
+        color_val = magnitude
+        if color_func is not None:
+            color_val = color_func(xx, yy, zz, uu, vv, ww)
+
         colors = None
         if colormap is not None:
-            colors = self.k3d.helpers.map_colors(
-                magnitude, colormap, [])
+            colors = self.k3d.helpers.map_colors(color_val, colormap, [])
 
         return origins, vectors, colors
 
@@ -463,6 +505,7 @@ class K3DBackend(Plot):
 
     def _update_interactive(self, params):
         np = import_module('numpy')
+        self._bounds = []
 
         # K3D title is an object in the figure
         n = len(self._fig.objects)
@@ -477,12 +520,15 @@ class K3DBackend(Plot):
                 self.series[i].params = params
 
                 if s.is_3Dline and s.is_point:
-                    x, y, z, _ = self.series[i].get_data()
+                    if s.is_parametric:
+                        x, y, z, _ = s.get_data()
+                    else:
+                        x, y, z = s.get_data()
                     positions = np.vstack([x, y, z]).T.astype(np.float32)
                     self._fig.objects[i].positions = positions
 
                 elif s.is_3Dline:
-                    x, y, z, _ = self.series[i].get_data()
+                    x, y, z, _ = s.get_data()
                     vertices = np.vstack([x, y, z]).T.astype(np.float32)
                     self._fig.objects[i].vertices = vertices
 
@@ -500,7 +546,8 @@ class K3DBackend(Plot):
                     self._fig.objects[i].vertices = vertices.T
                     if s.use_cm:
                         self._fig.objects[i].attribute = attribute
-                        self._fig.objects[i].color_range = [attribute.min(), attribute.max()]
+                        self._fig.objects[i].color_range = [float(attribute.min()), float(attribute.max())]
+                    self._high_aspect_ratio(x, y, z)
 
                 elif s.is_vector and s.is_3D:
                     if s.is_streamlines:
@@ -508,7 +555,8 @@ class K3DBackend(Plot):
 
                     xx, yy, zz, uu, vv, ww = self.series[i].get_data()
                     qkw, colormap = self._handles[i]
-                    origins, vectors, colors = self._build_k3d_vector_data(xx, yy, zz, uu, vv, ww, qkw, colormap)
+                    origins, vectors, colors = self._build_k3d_vector_data(xx, yy, zz, uu, vv, ww, qkw, colormap, s.normalize,
+                    s.color_func)
                     if colors is not None:
                         vec_colors = self._create_vector_colors(colors)
                         self.fig.objects[i].colors = vec_colors
@@ -526,17 +574,14 @@ class K3DBackend(Plot):
                         colors = colors.reshape((-1, 3))
                         colors = [self._rgb_to_int(c) for c in colors]
                         self._fig.objects[i].colors = colors
+                    self._high_aspect_ratio(x, y, z)
 
+        self._new_camera_position()
+        self._add_clipping_planes()
         # self._fig.auto_rendering = True
 
-    def show(self):
-        """Visualize the plot on the screen."""
+    def _new_camera_position(self):
         np = import_module('numpy')
-
-        if len(self._fig.objects) != len(self.series):
-            self._process_series(self._series)
-        self.plot_shown = True
-
         if len(self._bounds) > 0:
             # when there are very high aspect ratio meshes, or when zlim has
             # been set, we compute a new camera position to improve user
@@ -548,12 +593,31 @@ class K3DBackend(Plot):
             ).flatten()
             self._fig.camera = self._fig.get_auto_camera(1.5, 40, 60, bounds)
 
+    def show(self):
+        """Visualize the plot on the screen."""
+        if len(self._fig.objects) != len(self.series):
+            self._process_series(self._series)
         self._fig.display()
-        if self.zlim:
-            self._fig.clipping_planes = [
-                [0, 0, 1, self.zlim[0]],
-                [0, 0, -1, self.zlim[1]],
-            ]
+
+    def _add_clipping_planes(self):
+        if len(self._fig.clipping_planes) == 0:
+            clipping_planes = []
+            if self.zlim:
+                clipping_planes += [
+                    [0, 0, 1, -self.zlim[0]],
+                    [0, 0, -1, self.zlim[1]],
+                ]
+            if self.xlim:
+                clipping_planes += [
+                    [1, 0, 0, -self.xlim[0]],
+                    [-1, 0, 0, self.xlim[1]],
+                ]
+            if self.ylim:
+                clipping_planes += [
+                    [0, 1, 0, -self.ylim[0]],
+                    [0, -1, 0, self.ylim[1]],
+                ]
+            self._fig.clipping_planes = clipping_planes
 
     def save(self, path, **kwargs):
         """Export the plot to a static picture or to an interactive html file.
@@ -581,22 +645,22 @@ class K3DBackend(Plot):
         .. [#fn2] https://github.com/msgpack/msgpack-python
 
         """
-        if not self.plot_shown:
-            raise ValueError(
-                "K3D-Jupyter requires the plot to be shown on the screen "
-                + "before saving it."
-            )
 
         ext = os.path.splitext(path)[1]
         if not ext:
             path += ".png"
 
-        if ext in [".html", "htm"]:
+        if ext in [".html", ".htm"]:
             with open(path, 'w') as f:
                 include_js = kwargs.pop("include_js", False)
                 self.fig.snapshot_include_js = include_js
                 f.write(self.fig.get_snapshot(**kwargs))
         elif ext == ".png":
+            if self._get_mode() != 0:
+                raise ValueError(
+                    "K3D-Jupyter requires the plot to be shown on the screen "
+                    + "before saving a png file.")
+
             @self._fig.yield_screenshots
             def _func():
                 self._fig.fetch_screenshot(**kwargs)
