@@ -19,7 +19,16 @@ class PlotlyBackend(Plot):
         Possible values:
 
         - ``"equal"``: sets equal spacing on the axis of a 2D plot.
-        - ``"cube"``, ``"auto"`` for 3D plots.
+        - For 3D plots:
+
+          * ``"cube"``: fix the ratio to be a cube
+          * ``"data"``: draw axes in proportion to the proportion of their
+            ranges
+          * ``"auto"``: automatically produce something that is well
+            proportioned using 'data' as the default.
+          * manually set the aspect ratio by providing a dictionary.
+            For example: ``dict(x=1, y=1, z=2)`` forces the z-axis to appear
+            twice as big as the other two.
 
     rendering_kw : dict, optional
         A dictionary of keywords/values which is passed to Matplotlib's plot
@@ -33,7 +42,7 @@ class PlotlyBackend(Plot):
         * Refer to [#fn14]_ to customize implicit surface plots.
         * Refer to [#fn5]_ to customize 2D quiver plots. Default to:
           ``dict( scale = 0.075 )``.
-        * Refer to [#fn6]_ to customize 2D cone plots. Default to:
+        * Refer to [#fn6]_ to customize 3D cone plots. Default to:
           ``dict( sizemode = "absolute", sizeref = 40 )``.
         * Refer to [#fn8]_ to customize 2D streamlines plots. Defaul to:
           ``dict( arrow_scale = 0.15 )``.
@@ -48,6 +57,35 @@ class PlotlyBackend(Plot):
         If True, apply a color map to the meshes/surface. If False, solid
         colors will be used instead. Default to True.
 
+    annotations : list, optional
+        A list of dictionaries specifying the type the markers required.
+        The keys in the dictionary should be equivalent to the arguments
+        of the Plotly's `graph_objects.Scatter` class. Refer to [#fn15]_
+        for more information.
+        This feature is experimental. It might get removed in the future.
+
+    markers : list, optional
+        A list of dictionaries specifying the type the markers required.
+        The keys in the dictionary should be equivalent to the arguments
+        of the Plotly's `graph_objects.Scatter` class. Refer to [#fn3]_
+        for more information.
+        This feature is experimental. It might get removed in the future.
+
+    rectangles : list, optional
+        A list of dictionaries specifying the dimensions of the
+        rectangles to be plotted. The keys in the dictionary should be
+        equivalent to the arguments of the Plotly's
+        `graph_objects.Figure.add_shape` function. Refer to [#fn16]_
+        for more information.
+        This feature is experimental. It might get removed in the future.
+
+    fill : dict, optional
+        A list of dictionaries specifying the type the markers required.
+        The keys in the dictionary should be equivalent to the arguments
+        of the Plotly's `graph_objects.Scatter` class. Refer to [#fn17]_
+        for more information.
+        This feature is experimental. It might get removed in the future.
+
     References
     ==========
     .. [#fn1] https://plotly.com/python/contour-plots/
@@ -62,6 +100,9 @@ class PlotlyBackend(Plot):
     .. [#fn10] https://plotly.com/python/templates/
     .. [#fn13] https://github.com/plotly/plotly.js/issues/5003
     .. [#fn14] https://plotly.com/python/3d-isosurface-plots/
+    .. [#fn15] https://plotly.com/python/text-and-annotations/
+    .. [#fn16] https://plotly.com/python/shapes/
+    .. [#fn17] https://plotly.com/python/filled-area-plots/
 
 
     Notes
@@ -82,11 +123,16 @@ class PlotlyBackend(Plot):
     """
 
     _library = "plotly"
+    _allowed_keys = Plot._allowed_keys + [
+        "markers", "annotations", "fill", "rectangles"]
 
     colorloop = []
     colormaps = []
     cyclic_colormaps = []
     quivers_colors = []
+    wireframe_color = "#000000"
+
+    scattergl_threshold = 2000
 
     # color bar spacing
     _cbs = 0.15
@@ -100,6 +146,7 @@ class PlotlyBackend(Plot):
         plotly = import_module(
             'plotly',
             import_kwargs={'fromlist': ['graph_objects', 'figure_factory']},
+            warn_not_installed=True,
             min_module_version='5.0.0')
         go = plotly.graph_objects
 
@@ -144,6 +191,7 @@ class PlotlyBackend(Plot):
         self._theme = kwargs.get("theme", cfg["plotly"]["theme"])
         self.grid = kwargs.get("grid", cfg["plotly"]["grid"])
         self._fig = go.Figure()
+        self._colorbar_counter = 0
 
     @property
     def fig(self):
@@ -184,19 +232,19 @@ class PlotlyBackend(Plot):
         quivers_colors = self.quivers_colors if not tb.quivers_colors else tb.quivers_colors
         self._qc = itertools.cycle(quivers_colors)
 
-    def _create_colorbar(self, k, label, sc=False):
+    def _create_colorbar(self, label, sc=False):
         """This method reduces code repetition.
 
         Parameters
         ==========
-            k : int
-                index of the current color bar
             label : str
                 Name to display besides the color bar
             sc : boolean
                 Scale Down the color bar to make room for the legend.
                 Default to False
         """
+        k = self._colorbar_counter
+        self._colorbar_counter += 1
         return dict(
             x=1 + self._cbs * k,
             title=label,
@@ -214,6 +262,11 @@ class PlotlyBackend(Plot):
             col = next(self._cl)
         return [[0, col], [1, col]]
 
+    def _scatter_class(self, go, n, polar=False):
+        if not polar:
+            return go.Scatter if n < self.scattergl_threshold else go.Scattergl
+        return go.Scatterpolar if n < self.scattergl_threshold else go.Scatterpolargl
+
     def _process_series(self, series):
         np = import_module('numpy')
         plotly = import_module(
@@ -226,17 +279,9 @@ class PlotlyBackend(Plot):
         merge = self.merge
         self._init_cyclers()
 
-        # if legend=True and both 3d lines and surfaces are shown, then hide the
-        # surfaces color bars and only shows line labels in the legend.
-        # TODO: can I show both colorbars and legends by scaling down the color
-        # bars?
-        show_3D_colorscales = True
-        show_2D_vectors = False
-        for s in series:
-            if s.is_3Dline:
-                show_3D_colorscales = False
-            if s.is_2Dvector:
-                show_2D_vectors = True
+        mix_3Dsurfaces_3Dlines = (any(s.is_3Dsurface for s in series) and
+            any(s.is_3Dline and s.show_in_legend for s in series))
+        show_2D_vectors = any(s.is_2Dvector for s in series)
 
         self._fig.data = []
 
@@ -265,7 +310,6 @@ class PlotlyBackend(Plot):
                             ),
                             size=6,
                             showscale=self.legend and s.use_cm,
-                            colorbar=self._create_colorbar(ii, s.get_label(self._use_latex), True),
                         ),
                         customdata=param,
                         hovertemplate=(
@@ -274,8 +318,19 @@ class PlotlyBackend(Plot):
                             else "x: %{x}<br />y: %{y}<br />Arg: %{customdata}"
                         ),
                     )
+                    if lkw["marker"]["showscale"]:
+                        # only add a colorbar if required.
+
+                        # TODO: when plotting many (14 or more) parametric
+                        # expressions, each one requiring a colorbar, it might
+                        # happens that the horizontal space required by all
+                        # colorbars is greater than the available figure width.
+                        # That raises a strange error.
+                        lkw["marker"]["colorbar"] = self._create_colorbar(s.get_label(self._use_latex), True)
+
                     kw = merge({}, lkw, s.rendering_kw)
-                    self._fig.add_trace(go.Scatter(x=x, y=y, **kw))
+                    cls = self._scatter_class(go, len(x))
+                    self._fig.add_trace(cls(x=x, y=y, **kw))
                 else:
                     x, y = s.get_data()
                     color = next(self._cl) if s.line_color is None else s.line_color
@@ -298,37 +353,82 @@ class PlotlyBackend(Plot):
                     kw = merge({}, lkw, s.rendering_kw)
                     if s.is_polar:
                         kw.setdefault("thetaunit", "radians")
-                        self._fig.add_trace(go.Scatterpolar(r=y, theta=x, **kw))
+                        cls = self._scatter_class(go, len(x), True)
+                        self._fig.add_trace(cls(r=y, theta=x, **kw))
                     else:
-                        self._fig.add_trace(go.Scatter(x=x, y=y, **kw))
+                        cls = self._scatter_class(go, len(x))
+                        self._fig.add_trace(cls(x=x, y=y, **kw))
             elif s.is_3Dline:
                 # NOTE: As a design choice, I decided to show the legend entry
                 # as well as the colorbar (if use_cm=True). Even though the
                 # legend entry shows the wrong color (black line), it is useful
                 # in order to hide/show a specific series whenever we are
                 # plotting multiple series.
-                x, y, z, param = s.get_data()
+                if s.is_parametric:
+                    x, y, z, param = s.get_data()
+                else:
+                    x, y, z = s.get_data()
+                    param = np.ones_like(x)
+
                 if not s.is_point:
                     lkw = dict(
                         name=s.get_label(self._use_latex),
                         mode="lines",
-                        line=dict(
-                            width=4,
-                            colorscale=(
-                                next(self._cm)
-                                if s.use_cm
+                        showlegend=s.show_in_legend,
+                    )
+                    if self.legend and s.use_cm:
+                        # only add a colorbar if required.
+
+                        # TODO: when plotting many (14 or more) parametric
+                        # expressions, each one requiring a colorbar, it might
+                        # happens that the horizontal space required by all
+                        # colorbars is greater than the available figure width.
+                        # That raises a strange error.
+                        lkw["line"] = dict(
+                            width = 4,
+                            colorbar = self._create_colorbar(s.get_label(self._use_latex), True),
+                            colorscale = (
+                                next(self._cm) if s.use_cm
                                 else self._solid_colorscale(s)
                             ),
-                            color=param,
-                            showscale=self.legend and s.use_cm,
-                            colorbar=self._create_colorbar(ii, s.get_label(self._use_latex), True),
-                        ),
-                    )
+                            color = param,
+                            showscale = True,
+                        )
+                    else:
+                        lkw["line"] = dict(
+                            width = 4,
+                            color = (
+                                (next(self._cl) if s.line_color is None
+                                else s.line_color) if s.show_in_legend
+                                else self.wireframe_color)
+                        )
                 else:
+                    color = next(self._cl) if s.line_color is None else s.line_color
                     lkw = dict(
                         name=s.get_label(self._use_latex),
-                        mode="markers",
-                        line_color=next(self._cl) if s.line_color is None else s.line_color)
+                        mode="markers")
+
+                    lkw["marker"] = dict(
+                        color=color if not s.use_cm else param,
+                        size=8,
+                        colorscale=next(self._cm) if s.use_cm else None,
+                        showscale = True if s.use_cm else False,
+                    )
+                    if s.use_cm:
+                        lkw["marker"]["colorbar"] = self._create_colorbar(s.get_label(self._use_latex), True)
+
+                    if not s.is_filled:
+                        # TODO: how to show a colorscale if is_point=True
+                        # and is_filled=False?
+                        lkw["marker"] = dict(
+                            color="#E5ECF6",
+                            line=dict(
+                                width=2,
+                                color=lkw["marker"]["color"],
+                                colorscale=lkw["marker"]["colorscale"],
+                            )
+                        )
+
                 kw = merge({}, lkw, s.rendering_kw)
                 self._fig.add_trace(go.Scatter3d(x=x, y=y, z=z, **kw))
 
@@ -346,8 +446,8 @@ class PlotlyBackend(Plot):
                 colormap = next(self._cm)
                 skw = dict(
                     name=s.get_label(self._use_latex),
-                    showscale=self.legend and show_3D_colorscales,
-                    colorbar=self._create_colorbar(ii, s.get_label(self._use_latex)),
+                    showscale=self.legend,
+                    colorbar=self._create_colorbar(s.get_label(self._use_latex), mix_3Dsurfaces_3Dlines),
                     colorscale=colormap if s.use_cm else colorscale,
                     surfacecolor=surfacecolor,
                     cmin=surfacecolor.min(),
@@ -388,19 +488,28 @@ class PlotlyBackend(Plot):
                 yy = yy[:, 0]
                 ckw = dict(
                     contours=dict(
-                        coloring=None,
-                        showlabels=False,
+                        coloring=None if s.is_filled else "lines",
+                        showlabels=False if s.is_filled else True,
                     ),
                     colorscale=next(self._cm),
-                    colorbar=self._create_colorbar(ii, s.get_label(self._use_latex), show_2D_vectors),
+                    colorbar=self._create_colorbar(s.get_label(self._use_latex), show_2D_vectors),
+                    showscale=True if s.is_filled else False,
+                    zmin=zz.min(), zmax=zz.max()
                 )
                 kw = merge({}, ckw, s.rendering_kw)
                 self._fig.add_trace(go.Contour(x=xx, y=yy, z=zz, **kw))
                 count += 1
 
             elif s.is_vector:
+                if s.color_func is not None:
+                    warnings.warn("PlotlyBackend doesn't support custom "
+                        "coloring of 2D/3D quivers or streamlines plots. "
+                        "`color_func` will not be used.")
                 if s.is_2Dvector:
                     xx, yy, uu, vv = s.get_data()
+                    if s.normalize:
+                        mag = np.sqrt(uu**2 + vv**2 )
+                        uu, vv = [t / mag for t in [uu, vv]]
                     # NOTE: currently, it is not possible to create
                     # quivers/streamlines with a color scale:
                     # https://community.plotly.com/t/how-to-make-python-quiver-with-colorscale/41028
@@ -432,7 +541,7 @@ class PlotlyBackend(Plot):
                             ),
                             sizeref=0.3,
                             showscale=self.legend and s.use_cm,
-                            colorbar=self._create_colorbar(ii, s.get_label(self._use_latex)),
+                            colorbar=self._create_colorbar(s.get_label(self._use_latex)),
                             starts=dict(
                                 x=seeds_points[:, 0],
                                 y=seeds_points[:, 1],
@@ -457,12 +566,21 @@ class PlotlyBackend(Plot):
                                 w=ww.flatten(),
                                 **kw))
                     else:
+                        mag = np.sqrt(uu**2 + vv**2 + ww**2)
+                        if s.normalize:
+                            # NOTE/TODO: as of Plotly 5.9.0, it is impossible
+                            # to set the color of cones. Hence, by applying the
+                            # normalization, all cones will have the same
+                            # color.
+                            uu, vv, ww = [t / mag for t in [uu, vv, ww]]
                         qkw = dict(
                             showscale=(not s.is_slice) or self.legend,
                             colorscale=next(self._cm),
                             sizemode="absolute",
                             sizeref=40,
-                            colorbar=self._create_colorbar(ii, s.get_label(self._use_latex)),
+                            colorbar=self._create_colorbar(s.get_label(self._use_latex)),
+                            cmin=mag.min(),
+                            cmax=mag.max(),
                         )
                         kw = merge({}, qkw, s.rendering_kw)
                         self._fig.add_trace(
@@ -600,6 +718,25 @@ class PlotlyBackend(Plot):
                 kw = merge({}, lkw, s.rendering_kw)
                 self._fig.add_trace(go.Scatter(x=x, y=y, **kw))
 
+            elif s.is_generic:
+                if s.type == "markers":
+                    kw = merge({}, {"line_color": next(self._cl)}, s.rendering_kw)
+                    self._fig.add_trace(go.Scatter(*s.args, **kw))
+                elif s.type == "annotations":
+                    kw = merge({}, {
+                        "line_color": next(self._cl),
+                        "mode": "text",
+                        }, s.rendering_kw)
+                    self._fig.add_trace(go.Scatter(*s.args, **kw))
+                elif s.type == "fill":
+                    kw = merge({}, {
+                        "line_color": next(self._cl),
+                        "fill": "tozeroy"
+                        }, s.rendering_kw)
+                    self._fig.add_trace(go.Scatter(*s.args, **kw))
+                elif s.type == "rectangles":
+                    self._fig.add_shape(*s.args, **s.rendering_kw)
+
             else:
                 raise NotImplementedError(
                     "{} is not supported by {}".format(type(s), type(self).__name__)
@@ -635,11 +772,16 @@ class PlotlyBackend(Plot):
                         self.fig.data[i]["theta"] = x
 
                 elif s.is_3Dline:
-                    x, y, z, param = s.get_data()
+                    if s.is_parametric:
+                        x, y, z, param = self.series[i].get_data()
+                    else:
+                        x, y, z = self.series[i].get_data()
+                        param = np.zeros_like(x)
                     self.fig.data[i]["x"] = x
                     self.fig.data[i]["y"] = y
                     self.fig.data[i]["z"] = z
-                    self.fig.data[i]["line"]["color"] = param
+                    if s.use_cm:
+                        self.fig.data[i]["line"]["color"] = param
 
                 elif s.is_3Dsurface and (not s.is_domain_coloring) and (not s.is_implicit):
                     if not s.is_parametric:
@@ -660,11 +802,20 @@ class PlotlyBackend(Plot):
                 elif s.is_contour and (not s.is_complex):
                     _, _, zz = s.get_data()
                     self.fig.data[i]["z"] = zz
+                    self.fig.data[i]["zmin"] = zz.min()
+                    self.fig.data[i]["zmax"] = zz.max()
 
                 elif s.is_vector and s.is_3D:
                     if s.is_streamlines:
                         raise NotImplementedError
                     x, y, z, u, v, w = self.series[i].get_data()
+                    if s.normalize:
+                            # NOTE/TODO: as of Plotly 5.9.0, it is impossible
+                            # to set the color of cones. Hence, by applying the
+                            # normalization, all cones will have the same
+                            # color.
+                            mag = np.sqrt(u**2 + v**2 + w**2)
+                            u, v, w = [t / mag for t in [u, v, w]]
                     self.fig.data[i]["x"] = x.flatten()
                     self.fig.data[i]["y"] = y.flatten()
                     self.fig.data[i]["z"] = z.flatten()
@@ -674,6 +825,9 @@ class PlotlyBackend(Plot):
 
                 elif s.is_vector:
                     x, y, u, v = self.series[i].get_data()
+                    if s.normalize:
+                        mag = np.sqrt(u**2 + v**2 )
+                        u, v = [t / mag for t in [u, v]]
                     if s.is_streamlines:
                         # TODO: iplot doesn't work with 2D streamlines.
                         raise NotImplementedError
@@ -780,7 +934,8 @@ class PlotlyBackend(Plot):
                     zeroline=self.grid,  # thick line at x=0
                     visible=self.grid,  # numbers below
                 ),
-                aspectmode=(self.aspect if self.aspect != "equal" else "auto"),
+                aspectmode=("manual" if isinstance(self.aspect, dict) else (self.aspect if self.aspect != "equal" else "auto")),
+                aspectratio=self.aspect if isinstance(self.aspect, dict) else None
             ),
         )
 
@@ -811,7 +966,7 @@ class PlotlyBackend(Plot):
             self.process_series()
 
         ext = os.path.splitext(path)[1]
-        if ext.lower() in [".html", ".html"]:
+        if ext.lower() in [".htm", ".html"]:
             self.fig.write_html(path, **kwargs)
         else:
             self._fig.write_image(path, **kwargs)
