@@ -37,7 +37,7 @@ def convert_colormap(cm, to, n=256):
 
     assert isinstance(to, str)
     to = to.lower()
-    assert to in ["matplotlib", "plotly", "k3d", "bokeh"]
+    assert to in ["matplotlib", "plotly", "k3d", "bokeh", "mayavi"]
     if not isinstance(cm, (str, list, tuple, np.ndarray, Colormap)):
         raise ValueError(
             "`cm` must be either:\n"
@@ -48,6 +48,16 @@ def convert_colormap(cm, to, n=256):
             + "4. an array of colors extracted from a matplotlib.colors.Colormap."
         )
 
+    if to == "mayavi":
+        # NOTE: Mayavi colormaps are based on look up tables.
+        # It is possible to modify a colormap after an object (mesh) has been
+        # created (see this example):
+        # https://docs.enthought.com/mayavi/mayavi/auto/example_custom_colormap.html
+        # However, it is not possible to pass a look up table to the colormap
+        # keyword argument of a Mayavi function. Hence, we cannot implement
+        # intercompatibility with other plotting libraries.
+        return cm
+
     r = []
     if to == "k3d":
         # K3D color maps are lists of the form:
@@ -457,7 +467,7 @@ def get_seeds_points(xx, yy, zz, uu, vv, ww, to_numpy=True, **kw):
         return seeds
 
 
-def compute_streamtubes(xx, yy, zz, uu, vv, ww, kwargs):
+def compute_streamtubes(xx, yy, zz, uu, vv, ww, kwargs, color_func=None, ):
     """ Leverage vtk to compute streamlines in a 3D vector field.
 
     Parameters
@@ -469,6 +479,10 @@ def compute_streamtubes(xx, yy, zz, uu, vv, ww, kwargs):
     kwargs : dict
         Keyword arguments passed to the backend.
 
+    color_func : callable
+        The color function to apply to streamlines. Default to None, which
+        will apply the magnitude of the vector field.
+
 
     Returns
     =======
@@ -557,7 +571,13 @@ def compute_streamtubes(xx, yy, zz, uu, vv, ww, kwargs):
         streamline.GetPoints().GetData())
     streamlines_velocity = numpy_support.vtk_to_numpy(
         streamline.GetPointData().GetArray("vector_field"))
-    streamlines_speed = np.linalg.norm(streamlines_velocity, axis=1)
+
+    if color_func is None:
+        streamlines_speed = np.linalg.norm(streamlines_velocity, axis=1)
+    else:
+        x, y, z = streamlines_points.T
+        u, v, w = streamlines_velocity.T
+        streamlines_speed = color_func(x, y, z, u, v, w)
 
     vtkLines = streamline.GetLines()
     vtkLines.InitTraversal()
